"""This file contains the constants used by the multiplexer."""

INCOMING_QUEUE_MAX_BYTES_CHANNEL = 1024 * 1024 * 10
INCOMING_QUEUE_LOW_WATERMARK = 1024 * 512 * 1
INCOMING_QUEUE_HIGH_WATERMARK = 1024 * 1024 * 4

OUTGOING_QUEUE_MAX_BYTES_CHANNEL = 1024 * 1024 * 12
OUTGOING_QUEUE_LOW_WATERMARK = 1024 * 1024 * 2
OUTGOING_QUEUE_HIGH_WATERMARK = 1024 * 1024 * 6

PEER_TCP_TIMEOUT = 90
